@@ -1,26 +1,30 @@
 /*
-   freedup -- search for duplicate files in one or more directory hierarchies
-   Copyright (C) 2007,2008 AN@freedup.org.
+This file is part of FreeDup.
 
-   This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
+    FreeDup -- search for duplicate files in one or more directory hierarchies
+    Copyright (C) 2007,2008 Andreas Neuper
 
-   This program is distributed in the hope that it will be useful,
-   but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+    FreeDup is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation, either version 3 of the License, or
+    (at your option) any later version.
+
+    FreeDup is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with FreeDup.  If not, see <http://www.gnu.org/licenses/>.
 
-   You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 9 Temple Place - Suite 330, Boston, MA 02111-1307,
-   USA.
 */
 /* freedup was written by AN <AN@freedup.org>  */
 #include "sha1.h"
 #include "freedup.h"
+#include "helper.h"
+#include "web.h"
 #include "config.h"
+#include <signal.h>
 #include <assert.h>
 #include <ctype.h>
 
@@ -29,10 +33,14 @@ char *program_name;
 static void printhashchoice(void);
 static void printxtrachoice(void);
 
+
 void version(void)
 {
+#ifndef S_SPLINT_S
     printf( "\n\tfreedup: Version " VERSION FLAVOUR " \2512007-2008 by AN@freedup.org.\n"
 		"\t sha1.c/h of sha-1.0.4 \2512001-2003 by Allan Saddi.\n\n");
+#endif
+    return;
 }
 
 /* SYNOPSIS */
@@ -50,7 +58,7 @@ void usage(void)
     printf( "-%c --%s <seconds> allow the modification time stamps to differ by maximum #.\n",
 	long_options[4].val,long_options[4].name);
     printf( "-%c --%s <env>  load or store command line as an environment. There exist:\n\t\t",
-	long_options[5].val,long_options[5].name); listconfigs("%s, "); puts("");
+	long_options[5].val,long_options[5].name); listconfigs("%s, "); myputs("");
     printf( "-%c --%s\t   requires the path-stripped file names to be equal.\n",
 	long_options[6].val,long_options[6].name);
     printf( "-%c --%s\t   requires groups of the files to be equal before linking.\n",
@@ -79,9 +87,11 @@ void usage(void)
 	long_options[18].val,long_options[18].name);
     printf( "-%c --%s\t   generate symlinks although some given paths are relative.\n",
 	long_options[19].val,long_options[19].name);
+#ifdef HASHSUM
     printf( "-%c --%s <type>  select hash method: ",
 	long_options[20].val,long_options[20].name);
     printhashchoice();
+#endif
     printf( "-%c --%s\t   restore the time stamp of the directory after linking.\n",
 	long_options[21].val,long_options[21].name);
     printf( "-%c --%s\t   requires users of the files to be equal before linking.\n",
@@ -132,8 +142,8 @@ void usage(void)
 	int usehash=0;	/* -# */
 	int extratyp=0; /* -x */
 	int nosymlinks=0;	/* -s */
-	char*findopta=NULL;	/* -o */
-	char*pwd=NULL;	/* -b */
+/*@null@*/char*findopta=NULL;	/* -o */
+/*@null@*/char*pwd=NULL;	/* -b */
 	int hashmethod=0;	/* -t */
 
 
@@ -202,6 +212,7 @@ int exit_status;
 frdinfo*file_info;
 
 /* The array pointer we use for referencing duplicates in interactive mode. */
+/*@null@*/
 dinfo*dups_info=NULL;
 
 #include "auto.h"
@@ -212,7 +223,7 @@ static void printxtrachoice(void)
     unsigned char c=' ';
     for(n=1;extra[n].type!=NULL;n++)
     {
-	    if(c==',') putchar(c); else c=',';
+	    CHECK4KOMMA(c);
 	    printf(" %s",extra[n].type);
      /* else perror("lstat() failed while reading file statistics"); */
     }
@@ -233,7 +244,7 @@ int selectextra(const char*type)
     }
     if( extra[method].type == NULL )
     {
-	if(showmode)
+	if(showmode!=0)
 	    fprintf(stderr,
 		"Cannot determine a extra comparison module called \"%s\".\n", type);
 	return 0;
@@ -248,7 +259,7 @@ int hashread=0;
 struct hashtabelle {
 	const char *type;
 	const char *program;
-	const int   length;
+	const size_t length;
 	const int   space1;
 	const int   space2;
 } hashme[] = {
@@ -284,7 +295,7 @@ static void printhashchoice(void)
 	{
 	    if( S_ISREG(xstat.st_mode) && ((S_IXOTH|S_IXGRP|S_IXUSR)&xstat.st_mode) )
 	    {
-		if(c==',') putchar(c); else c=',';
+		CHECK4KOMMA(c);
 		printf("%s",hashme[n].type);
 	    }
 	}
@@ -294,7 +305,7 @@ static void printhashchoice(void)
     return;
 }
 
-inline int hashed(const char*s)
+int hashed(const char*s)
 {
     int returnval=atoi(s);
 
@@ -310,7 +321,7 @@ static void gethash_internal(frdinfo*a)
     size_t size,rdsz=0;
 
     file2hash = fopen (a->name, "rb");
-    if(file2hash == 0)
+    if(file2hash == NULL)
     {
 	/* perror("fopen() failed - file was probably deleted"); */
 	return;
@@ -328,7 +339,7 @@ static void gethash_internal(frdinfo*a)
     do {
         if( ferror(file2hash) && ! feof(file2hash) )
         {
-	    fclose(file2hash);
+	    if( fclose(file2hash) ) perror("Error closing file2hash file.");
 	    return;
         }
         SHA1UPDATE(&(a->context), buffer, rdsz);
@@ -336,7 +347,7 @@ static void gethash_internal(frdinfo*a)
         rdsz = fread(buffer, 1, (size<BLOCK)?size:BLOCK, file2hash);
         size = size - rdsz;
     } while( (!feof(file2hash))&&(rdsz!=0) );
-    fclose(file2hash);
+    if( fclose(file2hash) ) perror("Error closing file2hash file.");
     a->hash = calloc(SHA1_HASH_SIZE+2,sizeof(char));
     SHA1FINAL(&(a->context), (unsigned char*)a->hash);
     a->sha1pos = -1;
@@ -359,7 +370,7 @@ static void gethash(frdinfo*a)
     }
     sprintf(buffer,"%s \"%s\" 2>/dev/null",hashme[hashmethod].program, a->name);
     file2hash = popen (buffer, "r");
-    if(file2hash == 0)
+    if(file2hash == NULL)
     {
 	perror(buffer);
 	exit(-1);
@@ -425,19 +436,21 @@ int hashprobe(const int method,const char*filename)
     sprintf(buffer,"%s \"%s\" 2>/dev/null",
 	hashme[method].program, filename);
     file2hash = popen (buffer, "r");
-    if(file2hash == 0)
+    if(file2hash == NULL)
     {
 	perror(buffer);
 	return(-3);
     }
     fgets(buffer,MAXPATHLENGTH+132,file2hash);
+/* tests showed that instead of the result, sometimes the command is returned */
+/* reason is unclear - is a delay missing ?, use showmode to track it */
     pclose(file2hash);
     if( strlen(buffer) < hashme[method].length )
 	return -4;
     if( buffer[hashme[method].space1] != ' ' 
     ||  buffer[hashme[method].space2] != HASHCHAR )
     {
-	if(showmode)
+	if(showmode!=0)
 	    fprintf(stderr,
 		"hash result is: \"%s\"\n"
 		"%s: expected output format does not match template."
@@ -461,7 +474,7 @@ int selecthash(const char*type)
     }
     if( hashme[hashmethod].program==NULL)
     {
-	if(showmode)
+	if(showmode!=0)
 	    fprintf(stderr,
 		"Cannot determine a hash function called \"%s\".\n", type);
 	hashmethod=-1;
@@ -472,7 +485,7 @@ int selecthash(const char*type)
 	return hashmethod;
     } else
     {
-	if(showmode)
+	if(showmode!=0)
 	    fprintf(stderr,
 		"Hash function \"%s\" oputput format not supported.\n",
 		hashme[hashmethod].type);
@@ -590,7 +603,7 @@ void linkfilesecure(const char* aname,const char* bname,int(*lnk)(const char *,
      * later link it ;
      * finally remove tmp
      */
-    if( !testmode )
+    if( testmode==0 )
     {
 	strcpy(tmpfilename,bname);
 	strcat(tmpfilename,".XXXXXX");
@@ -600,28 +613,30 @@ void linkfilesecure(const char* aname,const char* bname,int(*lnk)(const char *,
 	 */
 	if( mktemp(tmpfilename) == NULL )
 	{
-	    perror("There is no unique temporory file name.");
+	    perror("There is no unique temporary file name.");
 	}
-	if( dirmtime )
+	if( dirmtime!=0 )
 	{
 	    dirname = calloc(sizeof(char),strlen(bname)+2);
 	    strcpy(dirname,bname);
 	    ptr=strrchr(dirname,'/');
 	    if( ptr!=NULL ) { *ptr='\0'; }
 	    gotdirtime = !lstat(dirname,&dstat);
-	    if( !gotdirtime )
+	    if( gotdirtime==0 )
 		perror("Cannot read directory time stamp.");
 	}
 	if( lstat(tmpfilename,&tstat) != 0 )
 	{
 	    /*
-	     * The errror needs not to be catched, since it is wanted
+	     * The error needs not to be catched, since it is wanted
 	     * that no file exists with the target name
 	     */
 	    rename( bname, tmpfilename );
 	} else
 	{
-	    fprintf(stderr,"Not able to rename \"%s\" to \"%s\".",bname,tmpfilename);
+	    // fprintf(stderr,"Not able to rename \"%s\" to \"%s\".",bname,tmpfilename);
+	    errno = EEXIST;
+	    perror("Not able to rename.");
 	}
 	if( lnk == symlink )
 	{
@@ -630,13 +645,28 @@ void linkfilesecure(const char* aname,const char* bname,int(*lnk)(const char *,
 	}
 	if( lnk( symaname, bname ) != 0 )
 	{
-	    perror("Linking failed.");
+	    // linking failed! 
+	    // try to move original in place again and log that fact
+	    //fprintf(stderr, "Linking failed. Trying rollback: \"%s\"", bname);
+	    errno = ECANCELED;
+	    perror("Linking failed - rollback appropriate");
+	    if ( rename( tmpfilename, bname ) != 0 )
+	    {
+		// moving old file in place again failed.
+		// check should be made to fail early, to pause or use noaction from now on
+		//fprintf(stderr, "unable to roll back: \"%s\" left", tmpfilename);
+		errno = ECANCELED;
+		perror("Rollback failed");
+	    }
 	}
-	if( unlink( tmpfilename ) != 0 )
-	{
-	    perror("Unlink failed.");
-	}
-	if( dirmtime & gotdirtime )
+	else 
+	    // unlink renamed original only of linking was successful
+	    if( unlink( tmpfilename ) != 0 )
+	    {
+		// unlinking failed!
+		fprintf(stderr, "Unlink failed: \"%s\"", tmpfilename);
+	    }
+	if( (dirmtime!=0) & (gotdirtime!=0) )
 	{
 	    utimecache.actime  = dstat.st_atime;
 	    utimecache.modtime = dstat.st_mtime;
@@ -645,7 +675,7 @@ void linkfilesecure(const char* aname,const char* bname,int(*lnk)(const char *,
 	    free( dirname );
 	}
     }
-    if( showmode || interact)
+    if( (showmode!=0) || (interact!=0) )
     {
 	printf("ln %s \"%s\" \"%s\"\n",( lnk == symlink )?"-s":"-f",symaname,bname);
     }
@@ -677,7 +707,7 @@ void linkfiles(const char* a, const char* b, const struct stat*ainfo)
 	 */
 	return;
     }
-    if(nosymlinks && (ainfo->st_dev != binfo->st_dev) )
+    if((nosymlinks!=0) && (ainfo->st_dev != binfo->st_dev) )
     {
         /* when on different devices and symlinks are not allowed
 	 * do not link, i.e. no also not prepare to link.
@@ -685,7 +715,7 @@ void linkfiles(const char* a, const char* b, const struct stat*ainfo)
 	 */
 	return;
     }
-    if(nonzero && (ainfo->st_size == 0) )
+    if((nonzero!=0) && (ainfo->st_size == 0) )
     {
         /* when only non-empty files are allowed to be linked
 	 * do not link, when they are identical and has size zero.
@@ -693,7 +723,7 @@ void linkfiles(const char* a, const char* b, const struct stat*ainfo)
 	 */
 	return;
     }
-    if( !weaklink && ( ainfo->st_dev == binfo->st_dev ) )
+    if((weaklink==0) && ( ainfo->st_dev == binfo->st_dev ) )
     {
 	/* 
 	 * either st_dev or st_ino are different,
@@ -703,12 +733,12 @@ void linkfiles(const char* a, const char* b, const struct stat*ainfo)
     } else
     {
 	assert( weaklink || !nosymlinks );
-	if(!hardlink )
+	if(hardlink==0 )
 	{
 	    linkfilesecure(a,b,symlink);
 	}
     }
-    if( showsave ) 
+    if( showsave!=0 ) 
     {
 	printf("Linking \"%s\" saves %d bytes\n",
 		basename(b), 
@@ -723,8 +753,8 @@ void realloc_dupinfo(dinfo*a)
 {
 	if(0) fprintf(stderr,"%s \t%d -> %d\n",a->name[a->count-1],(int)a->size, (int)(a->size)*2 );
 	a->size+= (a->size); /* double size */
-	a->name = realloc(a->name, sizeof(char**) * ( a->size ) );
-	a->prio = realloc(a->prio, sizeof(int*) * ( a->size ) );
+	a->name = realloc(a->name, sizeof(char**) * ( a->size + 1 ) );
+	a->prio = realloc(a->prio, sizeof(int*) * ( a->size  + 1) );
 	if( (a->name) == NULL || (a->prio) == NULL )
 	{
 	    perror("realloc() to extend dup info structure");
@@ -744,8 +774,8 @@ int build_dupinfo_list(frdinfo*a, frdinfo*b)
     {
 	assert(LETTERS>3);	 /* LETTERS should be 26 or greater for production */
 	dups_info[dupscount].size = LETTERS; /* originally designed for interaction */
-	dups_info[dupscount].name = (char**)calloc(dups_info[dupscount].size,sizeof(char**));
-	dups_info[dupscount].prio = (int*)calloc(dups_info[dupscount].size,sizeof(int*));
+	dups_info[dupscount].name = (char**)calloc(dups_info[dupscount].size+1,sizeof(char**));
+	dups_info[dupscount].prio = (int*)calloc(dups_info[dupscount].size+1,sizeof(int*));
 	dups_info[dupscount].name[0] = a->name;
 	dups_info[dupscount].prio[0] = a->prio;
 	dups_info[dupscount].name[1] = b->name;
@@ -760,11 +790,18 @@ int build_dupinfo_list(frdinfo*a, frdinfo*b)
     {
 	if( a->ident == NULL )
 	{
+
 	    /*clean up tangling references */
 	    if( (b->ident->replace!=NULL) )
 	    {
+		if(!b->ident->replace->name)
+		    b->ident->replace->name = b->ident->name;
 		b->ident = b->ident->replace;
 	    }
+	    if( b->ident->name==NULL ) {
+		b->ident->name = calloc(b->ident->size+1, sizeof(char**));
+	    }
+ 	    a->ident = b->ident;
 	    /* simply inherit reference list and append the new filename */
 	    a->ident = b->ident;
 	    for(n=0;b->ident->name[n]!=NULL;n++)
@@ -781,7 +818,12 @@ int build_dupinfo_list(frdinfo*a, frdinfo*b)
 	    /*clean up tangling references */
 	    if( (a->ident->replace!=NULL) )
 	    {
-		a->ident = a->ident->replace;
+		if(!a->ident->replace->name)
+		    a->ident->replace->name = a->ident->name;
+ 		a->ident = a->ident->replace;
+ 	    }
+	    if( a->ident->name==NULL ) {
+		a->ident->name = calloc(a->ident->size+1, sizeof(char**));
 	    }
 	    /* simply inherit reference list and append the new filename */
 	    b->ident = a->ident;
@@ -910,17 +952,17 @@ void dupinfo_global(dinfo*a,const char key)
 	case '#':  source=topsource; break;
 	case '<':  source=oldsource; break;
 	case '>':  source=newsource; break;
-	case '+':  if(extratyp) source=bigsource; else source=newsource; break;
-	case '-':  if(extratyp) source=tinysource; else source=oldsource; break;
+	case '+':  if(extratyp!=0) source=bigsource; else source=newsource; break;
+	case '-':  if(extratyp!=0) source=tinysource; else source=oldsource; break;
 	default :  source=0;         break;
     }
     lstat(a->name[source], sstat);
     printsource=totalinked;
     for( n=0; a->name[n]!=NULL; n++ )
     {
-	if( testmode )
+	if( testmode!=0 )
 	{
-	    if(showhard && n==0) puts("");
+	    if((showhard!=0) && n==0) myputs("");
 	    if( (lstat( a->name[n], astat)==0))
 	    {
 		if( n!=source
@@ -933,7 +975,7 @@ void dupinfo_global(dinfo*a,const char key)
 			printf( "\n%s\n", a->name[source] );
 		    printf( "%s\n", a->name[n] );
 		} else
-	        if(showhard)
+	        if(showhard!=0)
 		    printf( "%s\n", a->name[n] );
 	    }
         } else
@@ -1005,7 +1047,7 @@ int dupinfo_entry(dinfo*a)
 	    maxlink=astat->st_nlink;
 	    bigest=tinyest=astat->st_size;
 	}
-	if( testmode )
+	if( testmode!=0 )
 	{
 	    printf( "%s\n", a->name[n] );
 	} else
@@ -1065,7 +1107,7 @@ int dupinfo_entry(dinfo*a)
 	    tinyest=astat->st_size;
 	}
     }
-    if( !testmode )
+    if( testmode==0 )
     {
         static char buffer[MAXPATHLENGTH],*target=NULL;
 	/* 
@@ -1143,7 +1185,7 @@ int dupinfo_entry(dinfo*a)
 	/* 
 	 * tell user what we will do
 	 */
-	puts("");
+	myputs("");
 	lstat(a->name[source],sstat);
 	for( n=0; a->name[n]!=NULL; n++ )
 	{
@@ -1199,13 +1241,13 @@ void print_dupinfo(void)
     {
 	if(dups_info[n].name!=NULL)
 	{
-	    if(testmode)
+	    if(testmode!=0)
 	    {	dupinfo_global(&dups_info[n],globalkey);
 	    } else
 	    {
-	        if( showhard || dupinfo_notlinked(&dups_info[n]) )
+	        if( (showhard!=0) || dupinfo_notlinked(&dups_info[n]) )
 		{
-		    if(extratyp)
+		    if(extratyp!=0)
 			printf("\nDel:Lnk [filesize:devc:i-node:perm:L:tag] <filename>\n");
 		    else
 			printf("\nDel:Lnk [dev:i-node:perm:L] <filename>\n");
@@ -1318,8 +1360,8 @@ int comparison_by_content_hash(const void*a,const  void*b)
 	((frdinfo*)b)->sha1pos = -1;
 	hashread++;
     }
-    fclose(afile);
-    fclose(bfile);
+    if( fclose(afile) ) perror("Error closing hash comparison file A.");
+    if( fclose(bfile) ) perror("Error closing hash comparison file B.");
     if( result!=0 )
     {
 	/* this return is delayed. In case two files differ in the last byte,
@@ -1365,21 +1407,37 @@ int comparison_by_content(const void*a,const  void*b)
     {
         aread = fread(ablock,1,(asize<BLOCK)?asize:BLOCK,afile);
         asize = asize - aread;
-	assert((aread!=0)||feof(afile)||(asize==0));
-        bread = fread(bblock,1,(bsize<BLOCK)?bsize:BLOCK,bfile);
-        bsize = bsize - bread;
-	assert((bread!=0)||feof(bfile)||(bsize==0));
-	assert(aread==bread||feof(afile)||feof(bfile));
-	result=memcmp(ablock,bblock,aread);
+	if(!((aread!=0)||feof(afile)||(asize==0)))
+	{
+	    // should not happen - indicates inconsistent filesystems (e.g. /sys)
+	    // rarely matches some kind of race condition, i.e. file modified during access
+	    fprintf(stderr, "WARNING: File \"%s\" not properly read.\n", (((frdinfo*)a)->name) );
+	    result=-1;
+	} else 
+	{
+	    bread = fread(bblock,1,(bsize<BLOCK)?bsize:BLOCK,bfile);
+	    bsize = bsize - bread;
+	    if(!((bread!=0)||feof(bfile)||(bsize==0)))
+	    {
+		// should not happen - indicates inconsistent filesystems (e.g. /sys)
+		// rarely matches some kind of race condition, i.e. file modified during access
+		fprintf(stderr, "WARNING: File \"%s\" not properly read.\n", (((frdinfo*)b)->name) );
+		result=1;
+	    } else
+	    {
+		assert(aread==bread||feof(afile)||feof(bfile));
+	        result=memcmp(ablock,bblock,aread);
+	    }
+	}
 	if( result!=0 )
 	{
-	    fclose(afile);
-	    fclose(bfile);
+	    if( fclose(afile) ) perror("Error closing comparison file A file.");
+	    if( fclose(bfile) ) perror("Error closing comparison file B file.");
 	    return(result);
         }
     }
-    fclose(afile);
-    fclose(bfile);
+    if( fclose(afile) ) perror("Error closing comparison file A file.");
+    if( fclose(bfile) ) perror("Error closing comparison file B file.");
     build_dupinfo_list(((frdinfo*)a),((frdinfo*)b));
     return(0);
 }
@@ -1535,17 +1593,17 @@ if(0)	    printf("%01d\t%01d\t",identicalcount,alreadylinked);
 
 void printstats(void)
 {
-    puts("");
-    if(deleted)
+    myputs("");
+    if(deleted!=0)
 	fprintf(stderr,"%d Files (%d Bytes) have been%s deleted\n",
-		    deleted,freedspace,testmode?" marked to be":"");
+		    deleted,freedspace,(testmode!=0)?" marked to be":"");
     fprintf(stderr,"%d file%s of %d %sreplaced by links.\nThe total size of replac%s files %s %lld bytes.\n",
-	totalinked,(totalinked==1)?"":"s",filecount,testmode?"will be ":"",
-	testmode?"eable":"ed", testmode?"is":"was", linksaved);
-//    if( testmode && interact && extratyp!=0 )
+	totalinked,(totalinked==1)?"":"s",filecount,(testmode!=0)?"will be ":"",
+	(testmode!=0)?"eable":"ed", (testmode!=0)?"is":"was", linksaved);
+//    if( (testmode!=0) && (interact!=0) && (extratyp!=0) )
 //	fprintf(stderr,"Byte count might not be accurate, for a reliable prospect do not use '-i'.\n");
 #ifdef HASHSUM
-    if( usehash )
+    if( usehash!=0 )
     fprintf(stderr,"%s hash algorithm had to read %d files to avoid %d file comparisons.\n",
 	hashme[hashmethod].type,hashread,hashreject);
 #endif /* HASHSUM */
@@ -1587,7 +1645,7 @@ int readdirectory(frdinfo**fin, int*total, const int filecount, char*dir)
 	/*
 	 *
 	 */
-	if (!quietrun)
+	if (quietrun==0)
 	{
 	    if( (progress+1)/SCALE != (progress)/SCALE )
 	        fprintf(stderr, "\rReading directory tree %c ",
@@ -1623,7 +1681,7 @@ int readdirectory(frdinfo**fin, int*total, const int filecount, char*dir)
 	} else
 	if( S_ISREG((*fin)[count].info.st_mode) )
 	{
-	    if( nonzero && ((*fin)[count].info.st_size==0) )
+	    if( (nonzero!=0) && ((*fin)[count].info.st_size==0) )
 	    {
 	        free((*fin)[count].name);
 	        continue;
@@ -1689,7 +1747,7 @@ int readtree( frdinfo**fin, int*total, const int filecount, FILE*in)
 	     * Drop symbolic links and devices here
 	     */
 	    ||  !S_ISREG((*fin)[count].info.st_mode)
-	    || (nonzero && ((*fin)[count].info.st_size==0)) )
+	    || ((nonzero!=0) && ((*fin)[count].info.st_size==0)) )
 	    {
 		free((*fin)[count].name);
 		count--;
@@ -1712,6 +1770,21 @@ int readtree( frdinfo**fin, int*total, const int filecount, FILE*in)
     return count;
 }
 
+void finalize(void)
+{
+    if(interact!=0)  print_dupinfo();
+    else	  exec_dupinfo();
+    if(quietrun==0) printstats();
+	          return;
+}
+
+void finalsig(int sig)
+{
+    signal(SIGINT,SIG_DFL);
+    fprintf(stderr,"Signal received!\nLinking files found so far.");
+    exit(0);
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -1729,7 +1802,7 @@ main (int argc, char *argv[])
    * Time is not used in current implementation
    */
   start_time = time (NULL);
-  if(!quietrun) printf ("cur_day_start = %s", ctime (&cur_day_start));
+  if(quietrun==0) printf ("cur_day_start = %s", ctime (&cur_day_start));
 #endif /* DEBUG */
     exit_status = 0;
 
@@ -1740,7 +1813,7 @@ main (int argc, char *argv[])
         return(-1);
     }
     filecount=0;
-    while ((c = getopt_long(argc, argv, "ab:cdD:e:fghHik:lm:no:pP:qst:TuvVwWx:#:0",long_options,NULL)) != EOF) 
+    while ((c = getopt_long(argc, argv, "ab:cdD:e:fghHik:lm:Nno:pP:qst:TuvVwWx:#:0",long_options,NULL)) != (char)EOF) 
     {
         switch (c)
 	{
@@ -1765,9 +1838,11 @@ main (int argc, char *argv[])
 	  case 'h': usage();                  exit(0);
 	  case 'H': showhard	= 1-showhard;	break;
 	  case 'i': interact	= 1-interact;	break;
+	  case 'I': /*INETD*  for(;webconfig(HOMEDIR "/" WEBPAGE););*/ break;
 	  case 'k': globalkey	= optarg[0];	break;
 	  case 'l': hardlink	= 1-hardlink;	break;
 	  case 'm': minbytes	= atoi(optarg);	break;
+	  case 'N': testmode	= 1;		break;
 	  case 'n': testmode	= 1-testmode;	break;
 	  case 'o': if(findopta!=NULL) free(findopta);
 		    findopta	= calloc( strlen(optarg)+2, sizeof(char) );
@@ -1778,17 +1853,21 @@ main (int argc, char *argv[])
 	  case 'q': quietrun	= 1-quietrun;
 		    showsave	= showmode = 0;	break;
 	  case 's': nosymlinks	= 1-nosymlinks;	break;
+#ifdef HASHSUM
 	  case 't': hashmethod	= selecthash(optarg);
 						break;
+#endif
 	  case 'T': dirmtime	= 1-dirmtime;	break;
 	  case 'u': sameuser	= 1-sameuser;	break;
 	  case 'v': showmode	= 1-showmode;	break;
 	  case 'V': version();                exit(0);
 	  case 'w': weaklink	= 1-weaklink;	break;
-	  case 'W': for(;webconfig(WEBPAGE););	break;
+	  case 'W': dolisten();			break;
 	  case 'x': extratyp	= selectextra(optarg);
 						break;
+#ifdef HASHSUM
 	  case '#': usehash	= hashed(optarg);break;
+#endif
 	  case '0': nonzero	= 1-nonzero;	break;
 	  default:
 		fprintf(stderr,"wrong option: \"%c\"\n",c);
@@ -1796,14 +1875,14 @@ main (int argc, char *argv[])
 		exit(-1);
 	}
     }
-    if( sameperm!=0 && permmask==0) permmask=07777;
+    if( (sameperm!=0) && (permmask==0)) permmask=07777;
     if( environment!=NULL )
     {
 	if(rdcfg==0) writeconfig(environment,argc-optind,&(argv[optind]));
 	// if(rdcfg==0) writeconfig(environment);
     }
   
-    if( usehash && hashmethod!=0 )
+    if( (usehash!=0) && (hashmethod!=0) )
     {
 	if( hashme[hashmethod].program != NULL )
 	{
@@ -1816,23 +1895,27 @@ main (int argc, char *argv[])
 	if( hashme[hashmethod].program == NULL )
 	{
 	    usehash=0;
-	    fprintf(stderr,"No working hashmethod found.\n--> Use of hash methods is disabled.\n\n");
+	    fprintf(stderr,
+		    "No working hashmethod found.\n"
+		    "--> Use of hash methods is disabled.\n\n");
 	}
     }
-    if( 0 && !testmode && extratyp!=0 ) {
-	fprintf(stderr,"Extra Style comparisons are not thouroughly tested yet.\n"
-	"You may loose header information, if you press <return>.\nUse CTRL-C to stop.");
+    if( 1==0 && (testmode==0) && (extratyp!=0) ) {
+	fprintf(stderr,
+		"Extra Style comparisons are not thouroughly tested yet.\n"
+		"You may loose header information, if you press <return>.\n"
+		"Use CTRL-C to stop.");
 	scanf("%s",stopornot); }
     /* 
      * Collect files to scan.
      * Read them and fill in stat now
      */
-    if( argc == optind && filecount==0 )
+    if( (argc == optind) && (filecount==0) )
     {	/*
 	 * Read file names (not dir names) from stdin
 	 * if no directories where read from environment
 	 */
-	if(!quietrun) fprintf(stderr,"Taking file names from stdin\n\n");
+	if(quietrun==0) fprintf(stderr,"Taking file names from stdin\n\n");
 	filecount = readtree( &file_info, &frdinfosize, 0, stdin );
 	/*
 	 * Now we will skip the following for-loop
@@ -1866,10 +1949,17 @@ main (int argc, char *argv[])
 	perror("alloc() does not provide info structure for duplicate files");
 	exit(-1);
     }
-    if(!quietrun) fprintf(stderr,"\n%d files to investigate\n\n",filecount);
-    qsort( file_info, filecount, sizeof(frdinfo), comparison_by_size );
-    if(interact)  print_dupinfo();
+    if(quietrun==0) fprintf(stderr,"\n%d files to investigate\n\n",filecount);
+    if ( atexit(finalize) != 0 )
+    {
+	fprintf(stderr, "\nCannot set finalize() as exit function\n");
+    } else
+    {
+	signal(SIGINT,finalsig);
+        qsort( file_info, filecount, sizeof(frdinfo), comparison_by_size );
+    }
+/*  if(interact!=0)  print_dupinfo();
     else	  exec_dupinfo();
-    if(!quietrun) printstats();
+    if(quietrun==0) printstats(); 	*/
     exit(exit_status);
 }
